const now = new Date('1980-03-18T04:15:00')

const util = require('util')
util.promisify = jest.fn().mockImplementation(func => func)

const fs = require('fs')
fs.readdir = jest.fn()
fs.stat = jest.fn()

const os = require('os')
os.hostname = jest.fn().mockImplementation(() => 'fakehost')

/**
 * Asynchronous mock function that returns a promise witch
 * resolves or rejects based on the arguments given.
 * @param {object} unnamed - Either error or result
 */
function AsyncMock ({ result, error }) {
  return new Promise((resolve, reject) => {
    if (error) reject(error)
    resolve(result)
  })
}

const mockSpawn = jest.fn()
const mockSend = jest.fn()

const PythonWorker = require('../../src/pythonWorker')
// Mock constructor for the PythonWorker
jest.mock('../../src/pythonWorker', () => {
  return function (args) {
    this.spawn = mockSpawn
    this.send = mockSend
  }
})

const md5 = require('md5')
const interfaces = require('../../src/interfaces')
const { resolvers, typeDefs, __test__ } = interfaces
// const child_process = require('child_process')
// child_process.addCRPair({
//   input: '{"type":"options"}\n',
//   output: "[{'option1':'value1'},{'option2':'value2'}]"
// })

describe('interfaces module', () => {
  it('exports resolvers and typeDefs', () => {
    expect(resolvers).toBeDefined()
    expect(typeDefs).toBeDefined()
    expect(resolvers).toHaveProperty('Query'),
    expect(resolvers).toHaveProperty('Mutation')
  })

  it('finds workers', async () => {
    // Check that all possible return values of fs.readdir are correctly handled.
    const { findWorkers } = __test__

    // 1. Check use case
    const files = ['test1_worker.py', 'test2_worker.py']
    fs.readdir.mockReturnValueOnce(files)
    await expect(findWorkers('python_workers')).resolves.toEqual(files)

    // 2. Check empty directory
    fs.readdir.mockReturnValueOnce([])
    await expect(findWorkers('/')).resolves.toEqual([])

    // 3. Check argument error
    await expect(findWorkers()).rejects.toThrow(
      'Directory argument must be a string.'
    )

    // 4. Check non-existing path
    fs.readdir.mockImplementation(() => {
      throw new Error(
        `ENOENT: no such file or directory, scandir '/DOESNTEXIST'`
      )
    })
    await expect(findWorkers('/DOESNTEXIST')).rejects.toThrow(
      `ENOENT: no such file or directory, scandir '/DOESNTEXIST'`
    )
  })

  it('finds options', async () => {
    const { getOptions } = __test__
    const options = [{ option1: 'value1' }]
    const workerProcess = new PythonWorker()

    // 1. Check the use case
    mockSend.mockReturnValueOnce(AsyncMock({ result: { data: options } }))
    await expect(getOptions(workerProcess)).resolves.toEqual(options)
    expect(mockSend).toHaveBeenCalledWith({ type: 'options' })

    // 2. Check empty option case
    mockSend.mockReturnValueOnce(AsyncMock({ result: { data: [] } }))
    await expect(getOptions(workerProcess)).resolves.toEqual([])
    expect(mockSend).toHaveBeenCalledWith({ type: 'options' })

    // 3. Check error case
    mockSend.mockReturnValueOnce(AsyncMock({ result: { error: 'Timeout' } }))
    await expect(getOptions(workerProcess)).rejects.toThrow('Timeout')

    // 4. Throw error if workerProcess doesn't have a "send" property.
    expect(getOptions()).rejects.toThrow(
      'workerProcess not configured properly.'
    )
  })

  it('creates an interface', async () => {
    const { createInterface } = __test__
    const testInterfaces = []
    mockSpawn.mockReset()
    mockSend.mockReset()

    // 1. Check the use case
    mockSpawn.mockReturnValueOnce(AsyncMock({ result: { data: 'ready!' } }))
    mockSend.mockReturnValueOnce(AsyncMock({ result: { data: [] } }))
    fs.stat.mockReturnValueOnce({ mtime: now, size: 1234 })
    await expect(
      createInterface('python_workers', 'test1_worker.py', [])
    ).resolves.toContainEqual(
      expect.objectContaining({
        id: md5(`test1python_workers/test1_worker.py1980-03-18T03:15:00.000Z`),
        interfaceName: 'test1',
        options: [],
        workerProcess: expect.objectContaining({
          send: expect.anything(),
          spawn: expect.anything()
        })
      })
    )
    expect(mockSpawn).toHaveBeenCalled()
    expect(mockSend).toHaveBeenCalledWith({ type: 'options' })

    // 2. Check that the update property is set for existing interfaces

    // 3. Check that existing interfaces are skipped

    // 4. Check argument error case
    await expect(createInterface()).rejects.toThrow(
      'Directory argument must be a string.'
    )

    // 5. Check missing file case
    fs.stat.mockImplementation(() => {
      throw new Error(
        `ENOENT: no such file or directory, scandir 'python_workers/testXXX_worker.py'`
      )
    })
    await expect(
      createInterface('python_workers', 'testXXX_worker.py', [])
    ).rejects.toThrow(
      `ENOENT: no such file or directory, scandir 'python_workers/testXXX_worker.py'`
    )
  })

  it('finds interfaces', async () => {
    const { findInterfaces } = __test__
    interfaces.findWorkers = jest.fn()
    interfaces.findWorkers.mockReturnValueOnce([])
    interfaces.createInterface = jest.fn()
    interfaces.createInterface.mockReturnValueOnce([])
    const resp = await findInterfaces()
    expect(resp).toBe([])
  })
  const { Query, Mutation, Interface, Connection } = resolvers
})
